#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>
#include <pnglite.h>
#include "image.h"

#define THREADS_PER_BLOCK 1024

// guassian filter parameters
#define SIGMA 2.0f
#define FILTER_WIDTH 21

// nearest approximation of pi as a float (single precision) value
#define PI 3.1415927410125732421875f

void err_fatal(const char *fmt, ...)
{
  va_list args;

  va_start(args, fmt);
  vfprintf(stderr, fmt, args);
  va_end(args);

  exit(1);
}

unsigned long utime(void)
{
  struct timeval tv;
  unsigned long result = 0;

  gettimeofday(&tv, NULL);
  result += (tv.tv_sec * 1000000);
  result += tv.tv_usec;

  return result;
}

// Access red, green, blue, and alpha component values
// in a 32-bit unsigned RGBA pixel value.
#define ALPHA(pixel)  ((pixel)>>24)
#define BLUE(pixel)   (((pixel)>>16)&0xFF)
#define GREEN(pixel)  (((pixel)>>8)&0xFF)
#define RED(pixel)    ((pixel)&0xFF)

// Encode a 32-bit unsigned RGBA value from individual
// red, green, blue, and alpha component values.
#define RGBA(r,g,b,a) ((((a) << 24)) | (((b) << 16)) | (((g) << 8)) | ((r)))

#define sq(x) ((x)*(x))

float *create_filter(void)
{
  float *filter = (float *) malloc(FILTER_WIDTH*FILTER_WIDTH*sizeof(float));
  for (int i = 0; i < FILTER_WIDTH; i++) {
    for (int j = 0; j < FILTER_WIDTH; j++) {
      float dx = j - (FILTER_WIDTH / 2);
      float dy = i - (FILTER_WIDTH / 2);
      float fac = (1.0 / (2.0 * PI * sq(SIGMA))) *
                                    exp(-(sq(dx) + sq(dy)) / (2 * sq(SIGMA)));
      filter[i*FILTER_WIDTH + j] = fac;
    }
  }
  return filter;
}

__global__ void cuBlur(uint32_t *in, float *filter, uint32_t *out, int w, int h)
{
  // pixel to compute
  int x = blockIdx.x * THREADS_PER_BLOCK + threadIdx.x;

  if (x < w) {

    int y = blockIdx.y;

    // index of pixel in in/out buffers
    int index = y*w + x;

    // determine the pixel color component values by using the
    // filter matrix to compute a weighted average of the values
    // of nearby pixels
    float result[3] = { 0, 0, 0 }; // computed rgb values
    float weight_sum = 0;

    // which filter value to use
    unsigned filter_index = 0;

    // do the computation
    for (int i = y - FILTER_WIDTH/2; i <= y + FILTER_WIDTH/2; i++) {
      for (int j = x - FILTER_WIDTH/2; j <= x + FILTER_WIDTH/2; j++) {
        if (i >= 0 && i < h && j >= 0 && j < w) {
          int index = i*w + j;
          uint32_t orig_pixel = in[index];
          float fac = filter[filter_index];
          result[0] += (RED(orig_pixel) * fac);
          result[1] += (GREEN(orig_pixel) * fac);
          result[2] += (BLUE(orig_pixel) * fac);
          weight_sum += fac;
        }
        filter_index++;
      }
    }

    // normalize and clamp the computed pixel color component values
    for (int i = 0; i < 3; i++) {
      // normalize
      result[i] /= weight_sum;
      // clamp to range 0..255
      if (result[i] < 0) {
        result[i] = 0;
      } else if (result[i] > 255) {
        result[i] = 255;
      }
    }

    // render final pixel into the out buffer
    uint32_t transformed_pixel = RGBA(
      (unsigned)result[0], (unsigned)result[1], (unsigned)result[2],
      ALPHA(in[index])
    );
    out[index] = transformed_pixel;
  }
}

void execute_blur_cuda(struct Image *img, struct Image *result_image, float *filter) {
  // prepare to execute CUDA kernel
  cudaDeviceProp prop;
  cudaGetDeviceProperties( &prop, 0 );
  printf( "%d GPU processors, %d max threads per block\n",
                prop.multiProcessorCount,
                prop.maxThreadsPerBlock );

  // allocate device input and output buffers
  const size_t bufSize = img->width * img->height * sizeof(uint32_t);
  uint32_t *dev_imgdata, *dev_imgdata_out;
  cudaMalloc( (void **) &dev_imgdata, bufSize );
  cudaMalloc( (void **) &dev_imgdata_out, bufSize );

  // copy image data to device
  cudaMemcpy( dev_imgdata, img->data, bufSize, cudaMemcpyHostToDevice );

  // allocate device filter buffer
  float *dev_filter;
  const size_t filter_size = sizeof(float) * FILTER_WIDTH * FILTER_WIDTH;
  cudaMalloc( (void **) &dev_filter, filter_size );

  // copy filter to device
  cudaMemcpy( (void **) dev_filter, filter, filter_size, cudaMemcpyHostToDevice );

  int grid_w = img->width / THREADS_PER_BLOCK;
  if (img->width % THREADS_PER_BLOCK != 0) {
    grid_w++;
  }
  int grid_h = img->height;

  dim3 grid(grid_w, grid_h);

  // invoke the kernel!
  cuBlur<<<grid, THREADS_PER_BLOCK>>>(
    dev_imgdata, dev_filter, dev_imgdata_out, img->width, img->height
  );

  // copy transformed image data from device
  cudaMemcpy( result_image->data, dev_imgdata_out, bufSize, cudaMemcpyDeviceToHost );
}

void transform_pixel(uint32_t *in, uint32_t *out, int x, int y, int w, int h, float *filter) {
  float result[3] = { 0, 0, 0 }; // computed rgb values
  float weight_sum = 0;

  unsigned filter_index = 0;

  for (int i = y - FILTER_WIDTH/2; i <= y + FILTER_WIDTH/2; i++) {
    for (int j = x - FILTER_WIDTH/2; j <= x + FILTER_WIDTH/2; j++) {
      if (i >= 0 && i < h && j >= 0 && j < w) {
        int index = i*w + j;
        uint32_t orig_pixel = in[index];
        float fac = filter[filter_index];
        result[0] += (RED(orig_pixel) * fac);
        result[1] += (GREEN(orig_pixel) * fac);
        result[2] += (BLUE(orig_pixel) * fac);
        weight_sum += fac;
      }
      filter_index++;
    }
  }

  for (int i = 0; i < 3; i++) {
    // normalize
    result[i] /= weight_sum;
    // clamp to range 0..255
    if (result[i] < 0) {
      result[i] = 0;
    } else if (result[i] > 255) {
      result[i] = 255;
    }
  }

  int index = y*w + x;

  uint32_t transformed_pixel = RGBA(
    (unsigned)result[0], (unsigned)result[1], (unsigned)result[2],
    ALPHA(in[index])
  );
  out[index] = transformed_pixel;
}

void execute_blur(struct Image *img, struct Image *out, float *filter) {
  for (unsigned i = 0; i < img->height; i++) {
    for (unsigned j = 0; j < img->width; j++) {
      transform_pixel(img->data, out->data, (int)j, (int)i,
        (int)img->width, (int)img->height, filter);
    }
  }
}

void usage(void) {
  err_fatal(
    "Usage: readImage <mode> <infile> <outfile>\n"
    "Mode is:\n"
    "   cpu    render result image using CPU\n"
    "   gpu    render result image using GPU\n"
  );
}

enum Mode {
  USE_CPU,
  USE_GPU,
};

int main(int argc, char **argv)
{
  if (argc != 4) {
    usage();
  }

  const char *mode = argv[1];
  const char *in_filename = argv[2];
  const char *out_filename = argv[3];

  int exec_mode = -1;
  if (strcmp(mode, "cpu") == 0) {
    exec_mode = USE_CPU;
  } else if (strcmp(mode, "gpu") == 0) {
    exec_mode = USE_GPU;
  } else {
    usage();
  }

  // read image data
  struct Image *img = img_read_png(in_filename);
  if (!img) {
    err_fatal("Could not open png file '%s'\n", in_filename);
  }

  // create filter
  float *filter = create_filter();

  // create output buffer
  struct Image *result_image = img_create(img->width, img->height);

  unsigned long start_ts = utime();

  switch (exec_mode) {
  case USE_CPU:
    execute_blur(img, result_image, filter);
    break;
  case USE_GPU:
    execute_blur_cuda(img, result_image, filter);
    break;
  }

  unsigned long end_ts = utime();

  unsigned duration_millis = (end_ts - start_ts) / 1000UL;

  printf("Computation completed in %u.%03u seconds\n", duration_millis/1000U, duration_millis%1000U);

  img_write_png(result_image, out_filename);

  return 0;
}
